.version-stable {
    background-color: #2c7fb8ff;
    color: #2c7fb8ff;
}

.version-alert a {
    text-decoration: underline;
    background-color: #2c7fb8ff;
    color: #ddd;
}

.version-alert {
    background-color: #ff7f0eff;
    cff7f0effolor: #fff;
    font-family: "Work Sans", sans-serif;
    font-weight: 600; font-size: 16px;
}

.btn-version {
    background: #2c7fb8ff;
    color: #ffffff;
}

.btn-version:hover, .btn-version:focus, .btn-version:active, .btn-version.active, .open > .dropdown-toggle.btn-version {
    background: #33a6cc;
}

.btn-version:active, .btn-version.active {
    background: #007299;
    box-shadow: none;
}

/*Additional customizations*/
h1, h2, .rst-content .toctree-wrapper p.caption, h3, h4, h5, h6, legend {
    font-family: "Montserrat", "Lato", "Open Sans", sans-serif;
}

footer p {
    font-size: 14px;
    margin-top: 0px;
    margin-bottom: 0px;
}

h1 {
    margin-top: 0.0em;
    margin-bottom: 0.5em;
}

h2 {
    font-size: 140%;
    margin-top: 1.5em;
    margin-bottom: 0.5em;
}


h3 {
    font-size: 120%;
    margin-top: 1.0em;
    margin-bottom: 0.5em;
}

h4 {
    font-style: italic;
    font-size: 100%;
    margin-top: 0.5em;
    margin-bottom: 0.5em;
}

/* Layout changes */
.bd-main .bd-content .bd-article-container {
  max-width: 100em;  /* default is 60em */
}

.bd-page-width {
  max-width: 100rem;  /* default is 88rem */
}

.bd-sidebar-primary {
  width: 20em;
}

.bd-sidebar-secondary {
  width: 30em !important;
}