"""
Version and license information.
"""

__all__ = ['__version__', '__versiondate__', '__license__']

__version__ = '2.3.2'
__versiondate__ = '2025-07-16'
__license__ = f'Starsim {__version__} ({__versiondate__}) — © 2023-2025 by IDM'
